"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PreAggregationPartitionRangeLoader = void 0;
const shared_1 = require("@cubejs-backend/shared");
const QueryCache_1 = require("./QueryCache");
const PreAggregations_1 = require("./PreAggregations");
const PreAggregationLoader_1 = require("./PreAggregationLoader");
const DEFAULT_TS_FORMAT = 'YYYY-MM-DDTHH:mm:ss.SSS';
class PreAggregationPartitionRangeLoader {
    driverFactory;
    logger;
    queryCache;
    preAggregations;
    preAggregation;
    preAggregationsTablesToTempTables;
    loadCache;
    options;
    /**
     * Determines whether current instance instantiated for a jobbed build query
     * (initialized by the /cubejs-system/v1/pre-aggregations/jobs endpoint) or
     * not.
     */
    isJob;
    waitForRenew;
    requestId;
    lambdaQuery;
    dataSource;
    compilerCacheFn;
    constructor(driverFactory, logger, queryCache, preAggregations, preAggregation, preAggregationsTablesToTempTables, loadCache, options = {
        maxPartitions: 10000,
        maxSourceRowLimit: 10000,
    }) {
        this.driverFactory = driverFactory;
        this.logger = logger;
        this.queryCache = queryCache;
        this.preAggregations = preAggregations;
        this.preAggregation = preAggregation;
        this.preAggregationsTablesToTempTables = preAggregationsTablesToTempTables;
        this.loadCache = loadCache;
        this.options = options;
        this.isJob = !!options.isJob;
        this.waitForRenew = options.waitForRenew;
        this.requestId = options.requestId;
        this.lambdaQuery = options.lambdaQuery;
        this.dataSource = preAggregation.dataSource;
        this.compilerCacheFn = options.compilerCacheFn || ((subKey, cacheFn) => cacheFn());
    }
    async loadRangeQuery(rangeQuery, partitionRange) {
        const [query, values, queryOptions] = rangeQuery;
        const invalidate = this.preAggregation.invalidateKeyQueries?.[0]
            ? this.preAggregation.invalidateKeyQueries[0].slice(0, 2)
            : false;
        return this.queryCache.cacheQueryResult(query, values, QueryCache_1.QueryCache.queryCacheKey({
            query,
            values: values,
            invalidate,
        }), 24 * 60 * 60, {
            renewalThreshold: this.queryCache.options.refreshKeyRenewalThreshold
                || queryOptions?.renewalThreshold || 24 * 60 * 60,
            waitForRenew: this.waitForRenew,
            priority: this.priority(10),
            requestId: this.requestId,
            dataSource: this.dataSource,
            useInMemory: true,
            external: queryOptions?.external,
            renewalKey: partitionRange ? await this.getInvalidationKeyValues(partitionRange) : null,
        });
    }
    getInvalidationKeyValues(range) {
        const partitionTableName = PreAggregationPartitionRangeLoader.partitionTableName(this.preAggregation.tableName, this.preAggregation.partitionGranularity, range);
        return Promise.all((this.preAggregation.invalidateKeyQueries || []).map((sqlQuery) => (this.loadCache.keyQueryResult(this.replacePartitionSqlAndParams(sqlQuery, range, partitionTableName), this.waitForRenew, this.priority(10)))));
    }
    priority(defaultValue) {
        return this.preAggregation.priority ?? defaultValue;
    }
    async replaceQueryBuildRangeParams(queryValues) {
        if (queryValues.find(p => p === shared_1.BUILD_RANGE_START_LOCAL || p === shared_1.BUILD_RANGE_END_LOCAL)) {
            const [buildRangeStart, buildRangeEnd] = await this.loadBuildRange(this.preAggregation.timestampFormat);
            return queryValues?.map(param => {
                if (param === shared_1.BUILD_RANGE_START_LOCAL) {
                    return buildRangeStart;
                }
                else if (param === shared_1.BUILD_RANGE_END_LOCAL) {
                    return buildRangeEnd;
                }
                else {
                    return param;
                }
            });
        }
        return null;
    }
    replacePartitionSqlAndParams(query, dateRange, partitionTableName) {
        const [sql, params, options] = query;
        const updateWindowToBoundary = options?.incremental && (0, shared_1.addSecondsToLocalTimestamp)(dateRange[1], this.preAggregation.timezone, options?.updateWindowSeconds || 0);
        return [sql.replace(this.preAggregation.tableName, partitionTableName), params?.map(param => {
                if (dateRange && param === shared_1.FROM_PARTITION_RANGE) {
                    // Timestamp is in local timezone, so we need to convert to utc with desired format
                    return (0, shared_1.localTimestampToUtc)(this.preAggregation.timezone, this.preAggregation.timestampFormat, dateRange[0]);
                }
                else if (dateRange && param === shared_1.TO_PARTITION_RANGE) {
                    return (0, shared_1.localTimestampToUtc)(this.preAggregation.timezone, this.preAggregation.timestampFormat, dateRange[1]);
                }
                else {
                    return param;
                }
            }), {
                ...options,
                renewalThreshold: options?.incremental && updateWindowToBoundary < new Date() ?
                    // if updateWindowToBoundary passed just moments ago we want to renew it earlier in case
                    // of server and db clock don't match
                    Math.min(Math.round((new Date().getTime() - updateWindowToBoundary.getTime()) / 1000), options?.renewalThresholdOutsideUpdateWindow) :
                    options?.renewalThreshold
            }];
    }
    partitionPreAggregationDescription(range, buildRange) {
        const partitionTableName = PreAggregationPartitionRangeLoader.partitionTableName(this.preAggregation.tableName, this.preAggregation.partitionGranularity, range);
        const [_, buildRangeEnd] = buildRange;
        const loadRange = [...range];
        const partitionInvalidateKeyQueries = this.preAggregation.partitionInvalidateKeyQueries || this.preAggregation.invalidateKeyQueries;
        // `partitionInvalidateKeyQueries = []` in case of real time
        if ((!partitionInvalidateKeyQueries || partitionInvalidateKeyQueries.length > 0) && buildRangeEnd < range[1]) {
            loadRange[1] = buildRangeEnd;
        }
        const sealAt = (0, shared_1.addSecondsToLocalTimestamp)(loadRange[1], this.preAggregation.timezone, this.preAggregation.updateWindowSeconds || 0).toISOString();
        return {
            ...this.preAggregation,
            tableName: partitionTableName,
            structureVersionLoadSql: this.preAggregation.loadSql &&
                this.replacePartitionSqlAndParams(this.preAggregation.loadSql, range, partitionTableName),
            loadSql: this.preAggregation.loadSql &&
                this.replacePartitionSqlAndParams(this.preAggregation.loadSql, loadRange, partitionTableName),
            sql: this.preAggregation.sql &&
                this.replacePartitionSqlAndParams(this.preAggregation.sql, loadRange, partitionTableName),
            invalidateKeyQueries: (this.preAggregation.invalidateKeyQueries || [])
                .map(q => this.replacePartitionSqlAndParams(q, range, partitionTableName)),
            partitionInvalidateKeyQueries: this.preAggregation.partitionInvalidateKeyQueries?.map(q => this.replacePartitionSqlAndParams(q, range, partitionTableName)),
            indexesSql: (this.preAggregation.indexesSql || [])
                .map(q => ({ ...q, sql: this.replacePartitionSqlAndParams(q.sql, range, partitionTableName) })),
            previewSql: this.preAggregation.previewSql &&
                this.replacePartitionSqlAndParams(this.preAggregation.previewSql, range, partitionTableName),
            buildRangeStart: loadRange[0],
            buildRangeEnd: loadRange[1],
            sealAt, // Used only for kSql pre aggregations
        };
    }
    async loadPreAggregations() {
        if (this.preAggregation.partitionGranularity && !this.preAggregation.expandedPartition) {
            const loadPreAggregationsByPartitionRanges = async ({ buildRange, partitionRanges }) => {
                const partitionLoaders = partitionRanges.map(range => new PreAggregationLoader_1.PreAggregationLoader(this.driverFactory, this.logger, this.queryCache, this.preAggregations, this.partitionPreAggregationDescription(range, buildRange), this.preAggregationsTablesToTempTables, this.loadCache, this.options));
                const resolveResults = await Promise.all(partitionLoaders.map(async (l, i) => {
                    const result = await l.loadPreAggregation(false);
                    return result && {
                        ...result,
                        partitionRange: partitionRanges[i]
                    };
                }));
                return { loadResults: resolveResults.filter(res => res !== null), partitionLoaders };
            };
            // eslint-disable-next-line prefer-const
            let loadResultAndLoaders = await loadPreAggregationsByPartitionRanges(await this.partitionRanges());
            if (this.options.externalRefresh && loadResultAndLoaders.loadResults.length === 0) {
                loadResultAndLoaders = await loadPreAggregationsByPartitionRanges(await this.partitionRanges(true));
                // In case there are no partitions ready at matched time dimension intersection then no data can be retrieved.
                // We need to provide any table so query can just execute successfully.
                if (loadResultAndLoaders.loadResults.length > 0) {
                    loadResultAndLoaders.loadResults = [loadResultAndLoaders.loadResults[loadResultAndLoaders.loadResults.length - 1]];
                }
            }
            if (this.options.externalRefresh && loadResultAndLoaders.loadResults.length === 0) {
                throw new Error(
                // eslint-disable-next-line no-use-before-define
                PreAggregations_1.PreAggregations.noPreAggregationPartitionsBuiltMessage(loadResultAndLoaders.partitionLoaders.map(p => p.preAggregation)));
            }
            let { loadResults } = loadResultAndLoaders;
            let lambdaTable;
            let emptyResult = false;
            if (this.preAggregation.rollupLambdaId) {
                if (this.lambdaQuery && loadResults.length > 0) {
                    const { buildRangeEnd, targetTableName } = loadResults[loadResults.length - 1];
                    const lambdaTypes = await this.loadCache.getTableColumnTypes(this.preAggregation, targetTableName);
                    lambdaTable = await this.downloadLambdaTable(buildRangeEnd, lambdaTypes);
                }
                const rollupLambdaResults = this.preAggregationsTablesToTempTables.filter(tempTableResult => tempTableResult[1].rollupLambdaId === this.preAggregation.rollupLambdaId);
                const filteredResults = loadResults.filter(r => (this.preAggregation.lastRollupLambda || (0, shared_1.reformatInIsoLocal)(r.buildRangeEnd) === (0, shared_1.reformatInIsoLocal)(r.partitionRange[1])) &&
                    rollupLambdaResults.every(result => !result[1].buildRangeEnd || (0, shared_1.reformatInIsoLocal)(result[1].buildRangeEnd) < (0, shared_1.reformatInIsoLocal)(r.partitionRange[0])));
                if (filteredResults.length === 0) {
                    emptyResult = true;
                    loadResults = [loadResults[loadResults.length - 1]];
                }
                else {
                    loadResults = filteredResults;
                }
            }
            const allTableTargetNames = loadResults.map(targetTableName => targetTableName.targetTableName);
            let lastUpdatedAt = (0, PreAggregations_1.getLastUpdatedAtTimestamp)(loadResults.map(r => r.lastUpdatedAt));
            if (lambdaTable) {
                allTableTargetNames.push(lambdaTable.name);
                lastUpdatedAt = Date.now();
            }
            const unionTargetTableName = allTableTargetNames
                .map(targetTableName => `SELECT * FROM ${targetTableName}${emptyResult ? ' WHERE 1 = 0' : ''}`)
                .join(' UNION ALL ');
            return {
                targetTableName: allTableTargetNames.length === 1 && !emptyResult ? allTableTargetNames[0] : `(${unionTargetTableName})`,
                refreshKeyValues: loadResults.map(t => t.refreshKeyValues),
                lastUpdatedAt,
                buildRangeEnd: !emptyResult && loadResults.length && loadResults[loadResults.length - 1].buildRangeEnd,
                lambdaTable,
                rollupLambdaId: this.preAggregation.rollupLambdaId,
            };
        }
        else {
            return new PreAggregationLoader_1.PreAggregationLoader(this.driverFactory, this.logger, this.queryCache, this.preAggregations, this.preAggregation, this.preAggregationsTablesToTempTables, this.loadCache, this.options).loadPreAggregation(true);
        }
    }
    /**
     * Downloads the lambda table from the source DB.
     */
    async downloadLambdaTable(fromDate, lambdaTypes) {
        const { sqlAndParams, cacheKeyQueries } = this.lambdaQuery;
        const [query, params] = sqlAndParams;
        const values = params.map((p) => {
            if (p === shared_1.FROM_PARTITION_RANGE) {
                return fromDate;
            }
            if (p === shared_1.MAX_SOURCE_ROW_LIMIT) {
                return this.options.maxSourceRowLimit;
            }
            return p;
        });
        const { data } = await this.queryCache.renewQuery(query, values, cacheKeyQueries, 60 * 60, [query, values], undefined, {
            requestId: this.requestId,
            skipRefreshKeyWaitForRenew: false,
            dataSource: this.dataSource,
            external: false,
            useCsvQuery: true,
            lambdaTypes,
        });
        if (data.rowCount === this.options.maxSourceRowLimit) {
            throw new Error(`The maximum number of source rows ${this.options.maxSourceRowLimit} was reached for ${this.preAggregation.preAggregationId}`);
        }
        return {
            name: `${PreAggregations_1.LAMBDA_TABLE_PREFIX}_${this.preAggregation.tableName.replace('.', '_')}`,
            columns: data.types,
            csvRows: data.csvRows,
        };
    }
    async partitionPreAggregations() {
        if (this.preAggregation.partitionGranularity && !this.preAggregation.expandedPartition) {
            const { buildRange, partitionRanges } = await this.partitionRanges();
            return this.compilerCacheFn(['partitions', JSON.stringify(buildRange)], () => partitionRanges.map(range => this.partitionPreAggregationDescription(range, buildRange)));
        }
        else {
            return [this.preAggregation];
        }
    }
    async partitionRanges(ignoreMatchedDateRange) {
        const buildRange = await this.loadBuildRange();
        // buildRange was localized in loadBuildRange()
        // preAggregation.matchedTimeDimensionDateRange is also localized
        // in BaseFilter->formatToDate()/formatFromDate()
        let dateRange = PreAggregationPartitionRangeLoader.intersectDateRanges(buildRange, ignoreMatchedDateRange ? undefined : this.preAggregation.matchedTimeDimensionDateRange);
        if (!dateRange) {
            // If there's no date range intersection between query data range and pre-aggregation build range
            // use last partition so outer query can receive expected table structure.
            dateRange = [buildRange[1], buildRange[1]];
        }
        const partitionRanges = this.compilerCacheFn(['timeSeries', this.preAggregation.partitionGranularity, JSON.stringify(dateRange), `${this.preAggregation.timestampPrecision}`], () => PreAggregationPartitionRangeLoader.timeSeries(this.preAggregation.partitionGranularity, dateRange, this.preAggregation.timestampPrecision));
        if (partitionRanges.length > this.options.maxPartitions) {
            throw new Error(`Pre-aggregation '${this.preAggregation.tableName}' requested to build ${partitionRanges.length} partitions which exceeds the maximum number of partitions per pre-aggregation of ${this.options.maxPartitions}`);
        }
        return { buildRange: dateRange, partitionRanges };
    }
    async loadBuildRange(timestampFormat = DEFAULT_TS_FORMAT) {
        const { preAggregationStartEndQueries } = this.preAggregation;
        const [startDate, endDate] = await Promise.all(preAggregationStartEndQueries.map(async (rangeQuery) => PreAggregationPartitionRangeLoader.extractDate(await this.loadRangeQuery(rangeQuery), this.preAggregation.timezone, timestampFormat)));
        if (!this.preAggregation.partitionGranularity) {
            return this.orNowIfEmpty([startDate, endDate]);
        }
        // startDate & endDate are `localized` here
        const wholeSeriesRanges = PreAggregationPartitionRangeLoader.timeSeries(this.preAggregation.partitionGranularity, this.orNowIfEmpty([startDate, endDate]), this.preAggregation.timestampPrecision);
        const [rangeStart, rangeEnd] = await Promise.all(preAggregationStartEndQueries.map(async (rangeQuery, i) => PreAggregationPartitionRangeLoader.extractDate(await this.loadRangeQuery(rangeQuery, i === 0 ? wholeSeriesRanges[0] : wholeSeriesRanges[wholeSeriesRanges.length - 1]), this.preAggregation.timezone, timestampFormat)));
        return this.orNowIfEmpty([rangeStart, rangeEnd]);
    }
    now() {
        return (0, shared_1.utcToLocalTimeZone)(this.preAggregation.timezone, DEFAULT_TS_FORMAT, new Date().toJSON().substring(0, 23));
    }
    orNowIfEmpty(dateRange) {
        if (!dateRange[0] && !dateRange[1]) {
            const now = this.now();
            return [now, now];
        }
        if (!dateRange[0]) {
            return [dateRange[1], dateRange[1]];
        }
        if (!dateRange[1]) {
            return [dateRange[0], dateRange[0]];
        }
        return dateRange;
    }
    static checkDataRangeType(range) {
        if (!range) {
            return;
        }
        if (range.length !== 2) {
            throw new Error(`Date range expected to be an array with 2 elements but ${range} found`);
        }
        if (typeof range[0] !== 'string' || typeof range[1] !== 'string') {
            throw new Error(`Date range expected to be a string array but ${range} found`);
        }
        if ((range[0].length !== 23 && range[0].length !== 26) || (range[1].length !== 23 && range[0].length !== 26)) {
            throw new Error(`Date range expected to be in ${DEFAULT_TS_FORMAT} format but ${range} found`);
        }
    }
    static intersectDateRanges(rangeA, rangeB) {
        PreAggregationPartitionRangeLoader.checkDataRangeType(rangeA);
        PreAggregationPartitionRangeLoader.checkDataRangeType(rangeB);
        if (!rangeB) {
            return rangeA;
        }
        if (!rangeA) {
            return rangeB;
        }
        const from = rangeA[0] > rangeB[0] ? rangeA[0] : rangeB[0];
        const to = rangeA[1] < rangeB[1] ? rangeA[1] : rangeB[1];
        if (from > to) {
            return null;
        }
        return [
            from,
            to,
        ];
    }
    static timeSeries(granularity, dateRange, timestampPrecision) {
        if (!dateRange) {
            return [];
        }
        return (0, shared_1.timeSeries)(granularity, dateRange, {
            timestampPrecision
        });
    }
    static partitionTableName(tableName, partitionGranularity, dateRange) {
        let dateLenCut;
        switch (partitionGranularity) {
            case 'hour':
                dateLenCut = 13;
                break;
            case 'minute':
                dateLenCut = 16;
                break;
            default:
                dateLenCut = 10;
                break;
        }
        const partitionSuffix = dateRange[0].substring(0, dateLenCut).replace(/[-T:]/g, '');
        return `${tableName}${partitionSuffix}`;
    }
    static inDbTimeZone(preAggregationDescription, timestamp) {
        return (0, shared_1.localTimestampToUtc)(preAggregationDescription.timezone, preAggregationDescription.timestampFormat, timestamp);
    }
    static extractDate(data, timezone, timestampFormat = DEFAULT_TS_FORMAT) {
        return (0, shared_1.parseLocalDate)(data, timezone, timestampFormat);
    }
    static FROM_PARTITION_RANGE = shared_1.FROM_PARTITION_RANGE;
    static TO_PARTITION_RANGE = shared_1.TO_PARTITION_RANGE;
}
exports.PreAggregationPartitionRangeLoader = PreAggregationPartitionRangeLoader;
//# sourceMappingURL=PreAggregationPartitionRangeLoader.js.map